﻿using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.DotProlog.Code;
using OpcLabs.DotProlog.Code.Extensions;
using OpcLabs.DotProlog.Execution;
using OpcLabs.DotProlog.Execution.Extensions;
using OpcLabs.DotProlog.Terms;

namespace DotPrologTest.Execution
{
    [TestClass]
    public class EngineTest
    {
        #region Static Private Methods

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void ConsultAnimals([NotNull] IProgram program)
        {
            IDatabase database = program.Database;

            // animal(cat).
            // animal(dog).
            database.AssertFact(new NTerm("animal") { "cat" });
            database.AssertFact(new NTerm("animal") { "dog" });
        }

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void ConsultAppend([NotNull] IProgram program)
        {
            IDatabase database = program.Database;

            // ReSharper disable InconsistentNaming

            // append([],X,X).
            {
                var X = new Variable();
                database.AssertFact(new NTerm("append") { "[]", X, X });
            }
            // append([T|Q1],X,[T|Q2]) :- append(Q1,X,Q2).
            {
                var Q1 = new Variable();
                var Q2 = new Variable();
                var T = new Variable();
                var X = new Variable();
                database.AssertRule(
                    new NTerm("append") { new NList(Q1) { T }, X, new NList(Q2) { T } },
                    new Term[] { new NTerm("append") { Q1, X, Q2 } });
            }

            // ReSharper restore InconsistentNaming
        }

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void ConsultGenealogy([NotNull] IProgram program)
        {
            IDatabase database = program.Database;

            // ReSharper disable InconsistentNaming

            // parent(terach,abraham).
            // parent(abraham,isaac).
            // parent(isaac,jacob).
            // parent(jacob,benjamin).
            database.AssertFact(new NTerm("parent") { "terach", "abraham" });
            database.AssertFact(new NTerm("parent") { "abraham", "isaac" });
            database.AssertFact(new NTerm("parent") { "isaac", "jacob" });
            database.AssertFact(new NTerm("parent") { "jacob", "benjamin" });

            // ancestor(X,Y) :- parent(X,Y).
            {
                var X = new Variable();
                var Y = new Variable();
                database.AssertRule(
                    new NTerm("ancestor") { X, Y }, 
                    new Term[] { new NTerm("parent") { X, Y } });
            }
            // ancestor(X,Z) :- parent(X,Y), ancestor(Y,Z).
            {
                var X = new Variable();
                var Y = new Variable();
                var Z = new Variable();
                database.AssertRule(
                    new NTerm("ancestor") { X, Z }, 
                    new Term[] { new NTerm("parent") { X, Y }, new NTerm("ancestor") { Y, Z } });
            }

            // ReSharper restore InconsistentNaming
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void TestMethod0()
        {
            using (IEngine engine = new Engine())
            {
                // ReSharper disable InconsistentNaming
                var A = new Variable();
                var B = new Variable();
                var variableNames = new Dictionary<string, Variable>
                {
                    {"A", A},
                    {"B", B},
                };
                Term goal = (new Term[]
                {
                    new NTerm("=") {A, 1},
                    new NTerm(";") {new NTerm("=") {B, 2}, new NTerm("=") {B, 3},}
                }).ToPrologGoal();
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod1()
        {
            using (IEngine engine = new Engine())
            {
                ConsultAnimals(engine.Program);

                // ?- animal(A).
                // ReSharper disable InconsistentNaming
                var A = new Variable();
                var variableNames = new Dictionary<string, Variable> {{"A", A}};
                Term goal = new NTerm("animal") {A};
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod2()
        {
            using (IEngine engine = new Engine())
            {
                ConsultGenealogy(engine.Program);

                // ?- ancestor(terach,X).
                // ReSharper disable InconsistentNaming
                var X = new Variable();
                var variableNames = new Dictionary<string, Variable> {{"X", X}};
                Term goal = new NTerm("ancestor") {"terach", X};
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod3()
        {
            using (IEngine engine = new Engine())
            {
                ConsultAppend(engine.Program);

                // ?- append([a,b,c],[d,e,f],X).
                // ReSharper disable InconsistentNaming
                var X = new Variable();
                var variableNames = new Dictionary<string, Variable> {{"X", X}};
                Term goal = new NTerm("append") {new NList {"a", "b", "c"}, new NList {"d", "e", "f"}, X};
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod4()
        {
            using (IEngine engine = new Engine())
            {
                // ?- true.
                // ReSharper disable InconsistentNaming
                var variableNames = new Dictionary<string, Variable>();
                Term goal = new NTerm("true");
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod5()
        {
            using (IEngine engine = new Engine())
            {
                // ?- true.
                // ReSharper disable InconsistentNaming
                var variableNames = new Dictionary<string, Variable>();
                Term goal = new NTerm("repeat");
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames, 100);
            }
        }

        [TestMethod]
        public void TestMethod6()
        {
            using (IEngine engine = new Engine())
            {
                // ?- true.
                // ReSharper disable InconsistentNaming
                var variableNames = new Dictionary<string, Variable>();
                Term goal = new NTerm("fail");
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goal);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod8()
        {
            using (IEngine engine = new Engine())
            {
                ConsultGenealogy(engine.Program);

                // ?- ancestor(terach,X).
                // ReSharper disable InconsistentNaming
                var X = new Variable();
                var variableNames = new Dictionary<string, Variable> {{"X", X}};
                var goalSequence = new Term[] {new NTerm("ancestor") {"terach", X}, ControlConstructAtoms.Cut};
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goalSequence);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        [TestMethod]
        public void TestMethod9()
        {
            using (IEngine engine = new Engine())
            {
                ConsultGenealogy(engine.Program);

                // ?- ancestor(terach,X).
                // ReSharper disable InconsistentNaming
                var X = new Variable();
                var Y = new Variable();
                var variableNames = new Dictionary<string, Variable>
                {
                    {"X", X},
                    {"Y", Y},
                };
                var goalSequence = new Term[] {new NTerm("ancestor") {"terach", X}, new NTerm("=") {X, Y}};
                // ReSharper restore InconsistentNaming

                IEnumerable<Solution> results = engine.Call(goalSequence);
                FrontEnd.PrintResults(results, variableNames);
            }
        }

        #endregion



        #region Private Methods

        #endregion
    }
}
