﻿
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using JetBrains.Annotations;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.BaseLib.Testing;

namespace OpcLabs.UnitTestLib
{
    public abstract class Tester
    {
        #region Public Types

        public class CreatorEntry<TAttribute> where TAttribute : Attribute
        {
            public CreatorEntry([NotNull] Func<object> creator, [NotNull] TAttribute attribute)
            {
                Creator = creator;
                Attribute = attribute;
            }

            [NotNull]
            public TAttribute Attribute { get; }

            [NotNull]
            public Func<object> Creator { get; }
        }

        #endregion



        #region Protected Constructors

        protected Tester()
        {
            OptionDictionary = new StringObjectDictionary();
            Parameters = "";
            TestCategories = new Collection<string>();
            TestMethodDictionary = new Dictionary<string, TestMethod>();
        }

        #endregion



        #region Static Protected Methods

        [NotNull]
        static protected Dictionary<string, Func<object>> GetCreators(
            [NotNull] Type factoryType, 
            [NotNull] Type objectType,
            bool allowSubclasses)
        {
            Debug.Assert(factoryType != null);
            Debug.Assert(objectType != null);

            var creators = new Dictionary<string, Func<object>>();
            // ReSharper disable AssignNullToNotNullAttribute
            IEnumerable<MemberInfo> dataMemberInfos = factoryType.GetDataMembers(BindingFlags.Static | BindingFlags.Public)
                .Where(memberInfo =>
                {
                    // We will not use non-browseable data members as instances. Allows some level of control, e.g. to 
                    // prevent duplicities for Unequal testing.
                    var browseableAttribute = memberInfo.GetCustomAttribute<BrowsableAttribute>(inherit: true);
                    if ((browseableAttribute != null) && !browseableAttribute.Browsable)
                        return false;

                    Type dataType = memberInfo.GetDataType();
                    return
                        (dataType == objectType) ||
                        // ReSharper disable once PossibleNullReferenceException
                        (allowSubclasses && (dataType.IsSubclassOf(objectType)));
                });
            // ReSharper restore AssignNullToNotNullAttribute
            foreach (MemberInfo memberInfo in dataMemberInfos)
            {
                Debug.Assert(memberInfo != null);
                MemberInfo localMemberInfo = memberInfo;
                object Creator() => localMemberInfo.GetValue(null, null);
                creators.AddNew(memberInfo.Name, Creator);
            }
            return creators;
        }

        [NotNull]
        static protected Dictionary<string, CreatorEntry<TFactoryTypeAttribute>> GetCreatorEntriesFromAttributes<TFactoryTypeAttribute>(
            [NotNull] Type objectType,
            [NotNull] IEnumerable<TFactoryTypeAttribute> factoryTypeAttributes,
            bool allowSubclasses)
            where TFactoryTypeAttribute : FactoryTypeAttribute
        {
            Debug.Assert(objectType != null);

            var creatorEntries = new Dictionary<string, CreatorEntry<TFactoryTypeAttribute>>();
            foreach (TFactoryTypeAttribute factoryTypeAttribute in factoryTypeAttributes)
            {
                Debug.Assert(factoryTypeAttribute != null);
                Type instancesType = Type.GetType(factoryTypeAttribute.ClassName, throwOnError: true);
                Debug.Assert(instancesType != null);

                Dictionary<string, Func<object>> creators = GetCreators(instancesType, objectType, allowSubclasses);
                foreach (KeyValuePair<string, Func<object>> pair in creators)
                {
                    string name = pair.Key;
                    Func<object> func = pair.Value;
                    Debug.Assert(name != null);
                    Debug.Assert(func != null);

                    if (factoryTypeAttribute.ExceptMemberNames.Contains(name))
                        continue;

                    creatorEntries.AddNew(name, new CreatorEntry<TFactoryTypeAttribute>(func, factoryTypeAttribute));
                }
            }
            return creatorEntries;
        }

        #endregion



        #region Public Properties

        [NotNull]
        public Dictionary<string, CreatorEntry<InstancesAttribute>> InstanceCreators
        {
            get
            {
                lock (_serialize)
                {
                    // ReSharper disable once ConvertIfStatementToNullCoalescingExpression
                    if (_instanceCreators == null)
                        _instanceCreators = GetInstanceCreators();
                    return _instanceCreators;
                }
            }
        }

        [NotNull]
        public string Name
        {
            get
            {
                string name = GetType().Name;
                Debug.Assert(name != null);
                return name;
            }
        }

        [NotNull]
        public StringObjectDictionary OptionDictionary
        {
            get;
            set;
        }

        [NotNull] 
        public string Parameters { get; set; }

        [NotNull]
        public Collection<string> TestCategories { get; }

        [NotNull]
        public IEnumerable<string> TestMethodNames => TestMethodDictionary.Keys;

        [NotNull]
        public IEnumerable<TestMethod> TestMethods => TestMethodDictionary.Values;

        [CanBeNull]
        public Type TypeUnderTest
        {
            get => _typeUnderTest;
            set
            {
                Type oldValue = _typeUnderTest;
                _typeUnderTest = value;
                if (value != oldValue)
                    OnTypeUnderTestChange(oldValue);
            }
        }

        #endregion



        #region Public Methods

        public void AddTestMethod([NotNull] TestMethod testMethod)
        {
            string name = testMethod.Name;
            if (TestMethodDictionary.ContainsKey(name))
                throw new ArgumentException($"Test method dictionary already contains a method named \"{name}\".");
            TestMethodDictionary.AddNew(name, testMethod);
        }

        public bool FindTestMethod([NotNull] string name, [CanBeNull] out TestMethod testMethod)
        {
            Debug.Assert(name != null);
            return TestMethodDictionary.TryGetValue(name, out testMethod);
        }

        public bool GetOption([NotNull] string name)
        {
            Debug.Assert(name != null);
            return GetOption(name, true);
        }

        [CanBeNull]
        public object GetOption([NotNull] string name, [CanBeNull] object defaultValue)
        {
            Debug.Assert(name != null);
            return OptionDictionary.TryGetValue(name, out object value) ? value : defaultValue;
        }

        [CanBeNull]
        public T GetOption<T>([NotNull] string name, [CanBeNull] T defaultValue)
        {
            Debug.Assert(name != null);
            return (T) GetOption(name, (object)defaultValue);
        }

        [NotNull]
        public TestMethod GetTestMethod([NotNull] string name)
        {
            Debug.Assert(name != null);
            TestMethod testMethod = TestMethodDictionary[name];
            Debug.Assert(testMethod != null);
            return testMethod;
        }

        /// <inheritdoc cref="object.ToString()"/>
        public override string ToString()
        {
            return Name;
        }

        #endregion



        #region Protected Methods

        [NotNull]
        protected object CreateInstance([NotNull] string instanceName)
        {
            Debug.Assert(instanceName != null);
            CreatorEntry<InstancesAttribute> creatorEntry = InstanceCreators[instanceName];
            Debug.Assert(creatorEntry != null);
            Func<object> creatorFunc = creatorEntry.Creator;
            Debug.Assert(creatorFunc != null);
            object instance = creatorFunc();
            Debug.Assert(instance != null);
            return instance;
        }

        protected virtual void Initialize([NotNull] Type typeUnderTest)
        {
            TestMethodDictionary.Clear();
        }

        #endregion



        #region Private Properties

        [NotNull]
        private Dictionary<string, TestMethod> TestMethodDictionary { get; }

        #endregion



        #region Private Methods

        [NotNull]
        private Dictionary<string, CreatorEntry<InstancesAttribute>> GetInstanceCreators()
        {
            Debug.Assert(TypeUnderTest != null);

            IEnumerable<InstancesAttribute> attributes = TypeUnderTest.GetCustomAttributesNotNull<InstancesAttribute>(inherit: false);
            return GetCreatorEntriesFromAttributes(TypeUnderTest, attributes, allowSubclasses: false);
        }

        // ReSharper disable UnusedParameter.Local
        private void OnTypeUnderTestChange(Type oldValue)
        // ReSharper restore UnusedParameter.Local
        {
            if (TypeUnderTest != null)
                Initialize(TypeUnderTest);
        }

        #endregion



        #region Private Fields

        [CanBeNull] 
        private Dictionary<string, CreatorEntry<InstancesAttribute>> _instanceCreators;

        [CanBeNull]
        private Type _typeUnderTest;

        [NotNull]
        private readonly object _serialize = new object();

        #endregion
    }
}
