﻿
using System;
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.BaseLib.Threading;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    public class DALiveTestEnvironment
    {
        #region Private Constructors

        private DALiveTestEnvironment()
        {
            CommonServer Classifier(ServerDescriptor serverDescriptor)
            {
                return AllServers.First(server =>
                    // ReSharper disable once PossibleNullReferenceException
                    server.AllServerDescriptors.Any(descriptor => descriptor == serverDescriptor));
            }

            _monitorLockPool = new MonitorLockPool<ServerDescriptor>(
                new ClassifyingEqualityComparer<ServerDescriptor, CommonServer>(
                    Classifier, 
                    ReferenceEqualityComparer<CommonServer>.Default));
        }

        #endregion



        #region Static Public Properties

        #endregion



        #region Static Public Methods

        static public bool IsSupportedOptimistic([NotNull] ServerDescriptor serverDescriptor) // move elsewhere!
        {
            return true;
        }
        
        #endregion



        #region Static Public Fields

        [NotNull]
        static public readonly DALiveTestEnvironment Current = new DALiveTestEnvironment();

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcFoundation_XmlDaSampleServer OpcFoundation_XmlDaSampleServer = new OpcFoundation_XmlDaSampleServer(DALiveTestDefinitions.Locations.OpcXmlTest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<ServerDescriptor[]> OpcFoundation_XmlDaSampleServer_OptimisticSupportedServerDescriptorDataRows
        {
            get
            {
                return OpcFoundation_XmlDaSampleServer.AllServerDescriptors
                    .Where(IsSupportedOptimistic)
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcLabs_KitServer OpcLabs_KitServer = new OpcLabs_KitServer(DALiveTestDefinitions.Locations.OpcComTest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<ServerDescriptor[]> OpcLabs_KitServer_OptimisticSupportedServerDescriptorDataRows
        {
            get
            {
                return OpcLabs_KitServer.AllServerDescriptors
                    .Where(IsSupportedOptimistic)
                    .Select(descriptor => new[] { descriptor });
            }
        }

        #endregion



        #region Public Properties

        [NotNull]
        public IEnumerable<CommonServer> AllServers
        {
            get
            {
                yield return OpcFoundation_XmlDaSampleServer;
            }
        }

        #endregion



        #region Public Methods

        public bool CanConnectServer([NotNull] ServerDescriptor serverDescriptor)
        {
            return TestServer(serverDescriptor) == 0;
        }

        public void PrecheckServer([NotNull] ServerDescriptor serverDescriptor)
        {
            AssertUtilities.Phase($"Pre-checking server \"{serverDescriptor}\"");

            int error = TestServer(serverDescriptor);
            if (error != 0)
            {
                string message = $"Server \"{serverDescriptor}\" pre-check failed: {error}";
                AssertUtilities.WriteLine(message);
                Assert.Inconclusive(message);
            }
        }

        [NotNull]
        public IDisposable RequestLockFor([NotNull] ServerDescriptor serverDescriptor)
        {
            return _monitorLockPool.RequestLockFor(serverDescriptor);
        }

        public int TestServer([NotNull] ServerDescriptor serverDescriptor)
        {
            int error;
            lock (_testServerCache)
            {
                if (!_testServerCache.TryGetValue(serverDescriptor, out error))
                {
                    error = InternalTestServer(serverDescriptor);
                    _testServerCache.AddNew(serverDescriptor, error);
                }
            }
            return error;
        }

        #endregion



        #region Public Fields

        #endregion



        #region Private Methods

        private int InternalTestServer([NotNull] ServerDescriptor serverDescriptor)
        {
            switch (serverDescriptor.Technology)
            {
                case OpcTechnologies.Com:
                    if (!string.IsNullOrEmpty(serverDescriptor.MachineName))
                        return 0;                    // TODO: remove this - is only because we cannot test remote servers yet
                    if (!serverDescriptor.CanInstantiate())
                        return -1;
                    return 0;

                /* TODO:

                string machineName = serverDescriptor.MachineName;
                if (!machineName.IsEmpty())
                {
                    PingReply pingReply;
                    try
                    {
                        pingReply = PingUtilities.Send(machineName);
                    }
                    catch (NotSupportedException)
                    {
                        // address is an IPv6 address and the local computer is running an operating system earlier than Windows 2000. 
                        return -2;
                    }
                    catch (PingException)
                    {
                        // An exception was thrown while sending or receiving the ICMP messages. See the inner exception for the exact 
                        // exception that was thrown.
                        return -1;
                    }
                    if (pingReply.Status != IPStatus.Success)
                        return -3;
                }

                // ServerDescriptor pre-check passes for all pingable appear-remote machines, regardless of the server class.
                bool appearsLocal =
                    machineName.IsEmpty() ||
                    HostNameUtilities.IsLocalHostOrAddress(machineName) ||
                    (machineName.ToUpper() == System.Environment.MachineName.ToUpper()) ||
                    HostNameUtilities.Match(machineName, System.Net.Dns.GetHostName());
                // Cannot check instantiation feasibility on remote machines.
                if (appearsLocal)
                {
                    ServerDescriptor localServerDescriptor = serverDescriptor.TypedClone();
                    localServerDescriptor.MachineName = "";
                    if (!localServerDescriptor.CanInstantiate())
                        return -4;
                }

                return 0;
                */

                case OpcTechnologies.Xml:
                    return TcpClientUtilities.TestCreate(serverDescriptor, 
                        millisecondsTimeout: 8000); // 5000 milliseconds was not enough under Linux VM and with VPN
            }

            return 0;
        }

        #endregion



        #region Private Fields

        [NotNull] 
        private readonly MonitorLockPool<ServerDescriptor> _monitorLockPool;

        [NotNull]
        private readonly Dictionary<ServerDescriptor, int> _testServerCache = new Dictionary<ServerDescriptor, int>();

        #endregion
    }
}
