﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    public partial class EasyDAClientTest
    {
        #region Static Public Methods

        static public void SubscribeMultipleItems_Incrementing([NotNull] string machineName, int itemCount)
        {
            TestSection.Arrange();

            ServerDescriptor serverDescriptor = new ServerDescriptor(machineName, "OPCLabs.KitServer.2");
            DALiveTestEnvironment.Current.PrecheckServer(serverDescriptor);

            var eventArgsList = new List<EasyDAItemChangedEventArgs>();
            bool acceptingNotifications = true;

            var argumentsArray = new EasyDAItemSubscriptionArguments[itemCount];
            const int phaseCount = 100;
            for (int i = 0; i < itemCount; i++)
            {
                int copy = i / phaseCount;
                int phase = i % phaseCount;
                string itemId = String.Format(CultureInfo.InvariantCulture,
                    "Simulation.Incrementing (10 s).Copy_{0}.Phase_{1}", copy + 1, phase + 1);

                argumentsArray[i] = new EasyDAItemSubscriptionArguments(
                    // ReSharper disable RedundantArgumentName
                    serverDescriptor: serverDescriptor,
                    itemDescriptor: itemId,
                    groupParameters: 60 * 1000,
                    state: i,
                    // ReSharper disable once AccessToModifiedClosure
                    callback: (sender, args) =>
                    {
                        if (acceptingNotifications) eventArgsList.Add(args);
                    });
                // ReSharper restore RedundantArgumentName
            }

            IEasyDAClient target = new EasyDAClient();

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            var stopwatch = new Stopwatch();
            stopwatch.Start();
            target.SubscribeMultipleItems(argumentsArray); // should not crash here
            stopwatch.Stop();
            AssertUtilities.WriteLine($"Elapsed: {stopwatch.Elapsed}");

            // Give time for notifications to arrive
            //AssertUtilities.WaitFor(5*1000 + (int)(itemCount*0.01)*1000);
            AssertUtilities.WaitingPhase(80 * 1000);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeAllItems();

            acceptingNotifications = false;
            AssertUtilities.WaitingPhase(2 * 1000);

            TestSection.Assert();

            int count = eventArgsList.Count;
            AssertUtilities.WriteLine($"Count: {count}");
            Assert.AreNotEqual(0, count);

            foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsList)
                AssertUtilities.WriteLine(eventArgs);
            AssertUtilities.WriteLine(".");

            AssertUtilities.BeginWriteLine();
            Trace.Write("Checking... ");
            for (int i = 0; i < count; i++)
            {
                EasyDAItemChangedEventArgs eventArgs = eventArgsList[i];
                Trace.Write(i + " ");

                Assert.IsNotNull(eventArgs);
                if (eventArgs.Exception != null)
                {
                    AssertUtilities.WriteLine(eventArgs.Arguments.ItemDescriptor);
                    AssertUtilities.WriteLine(eventArgs.Exception);
                }
                Assert.IsNull(eventArgs.Exception);
                // IMPROVE: test more
            }
            AssertUtilities.EndWriteLine();
            Assert.IsTrue(count >= itemCount);

            // IMPROVE: test what really happens and how it works internally...
        }

        static public void SubscribeMultipleItems_UnknownItems(int itemCount)
        {
            const string serverClass = "OPCLabs.KitServer.2";
            const int unknownItemId =
                unchecked((int)0xC0040007 /*The item is no longer available in the server address space.*/);

            var eventArgsList = new List<EasyDAItemChangedEventArgs>();

            int lastNotificationTick = -1;
            var argumentsArray = new EasyDAItemSubscriptionArguments[itemCount];
            for (int i = 0; i < itemCount; i++)
            {
                argumentsArray[i] = new EasyDAItemSubscriptionArguments(
                    // ReSharper disable RedundantArgumentName
                    machineName: "",
                    serverClass: serverClass,
                    itemId: String.Format(CultureInfo.InvariantCulture, "BFLMPSVZ.UnknownItem_{0}", i),
                    requestedUpdateRate: 10 * 1000,
                    state: i,
                    callback: (sender, args) =>
                    {
                        lastNotificationTick = System.Environment.TickCount;
                        eventArgsList.Add(args);
                    });
            }
            // ReSharper restore RedundantArgumentName

            IEasyDAClient target = new EasyDAClient();

            AssertUtilities.Phase("Subscribing");
            var stopwatch = new Stopwatch();
            stopwatch.Start();
            target.SubscribeMultipleItems(argumentsArray); // should not crash here
            stopwatch.Stop();
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");
            AssertUtilities.WriteLine($"Elapsed: {stopwatch.Elapsed}");

            // Give time for notifications to arrive
            AssertUtilities.WaitingPhase(20 * 1000);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeAllItems();
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");

            AssertUtilities.WaitingPhase(5 * 1000);
            AssertUtilities.WriteLine($"Tick count: {System.Environment.TickCount}");

            int notificationCount = eventArgsList.Count;
            AssertUtilities.WriteLine($"Notification count: {notificationCount}");
            AssertUtilities.WriteLine($"Last notification tick: {lastNotificationTick}");
            Assert.AreNotEqual(0, notificationCount);

            foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsList)
                AssertUtilities.WriteLine(eventArgs);
            AssertUtilities.WriteLine(".");

            AssertUtilities.BeginWriteLine();
            Trace.Write("Checking... ");
            for (int i = 0; i < notificationCount; i++)
            {
                EasyDAItemChangedEventArgs eventArgs = eventArgsList[i];
                Trace.Write(i + " ");

                Assert.IsNotNull(eventArgs);
                Exception exception = eventArgs.Exception;
                Assert.IsNotNull(exception);

                Assert.AreEqual(unknownItemId, exception.HResult);
                Assert.AreEqual(serverClass, exception.Source);

                if (exception is COMException comException)
                {
                    Assert.AreEqual(unknownItemId, comException.ErrorCode);
                    continue;
                }

                if (exception is NormalizedException)
                {
                    continue;
                }

                Assert.Fail($"Incorrect exception type: {exception.GetType()}");
            }
            AssertUtilities.EndWriteLine();
            Assert.AreEqual(itemCount, notificationCount);

            // IMPROVE: test what really happens and how it works internally...
        }

        #endregion
    }
}
