﻿
using System.Diagnostics;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.IdentityModel.User;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.UA;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Test Methods

        [TestMethod]
        public void Group001_Session_Header____________________()
        {
        }

        /// <summary>
        /// CreateSession() over an insecure channel with UA TCP and SecurityPolicy =None
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test01_CreateSession_UaTcp_None()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
                { EndpointSelectionPolicy = new UAEndpointSelectionPolicy(UAMessageSecurityModes.SecurityNone) };

            TestSection.Act();
            using (var client = new EasyUAClient {Isolated = true})
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// CreateSession() over a secure channel with UA TCP and SecurityPolicy=Basic128Rsa15 where the key length is minimum 2048.
        /// </summary>
        // IMPROVE: we are not yet imposing the 2048 MinimumKeySize. Can be done in CertificateAcceptancePolicy, but that is global.
        [TestMethod]
        public void Group001_Session_Test02_CreateSession_UaTcp_Basic128Rsa15()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
            {
                EndpointSelectionPolicy = new UAEndpointSelectionPolicy
                    { SecurityPolicyUriString = UASecurityPolicyUriStrings.Basic128Rsa15 }
            };

            TestSection.Act();
            using (var client = new EasyUAClient {Isolated = true})
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// ActivateSession() with user identity = username/password
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test03_ActivateSession_UserNameAndPassword()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            string userName = UserName;
            Debug.Assert(userName != null, "Missing user name");
            var endpointDescriptor = new UAEndpointDescriptor(EndpointDescriptorUserNameAndPassword)
                { UserIdentity = UserIdentity.CreateUserNameIdentity(userName, Password) };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// ActivateSession() with user identity = anonymous
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test04_ActivateSession_UserIdentityAnonymous()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            var endpointDescriptor = new UAEndpointDescriptor(EndpointDescriptorCommon)
                { UserIdentity = UserIdentity.CreateAnonymousIdentity() };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// ActivateSession - change user in an existing session.
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test05_ActivateSession_ChangeUser()
        {
            ClientNotSupported();
            // IMPROVE: have this
        }

        /// <summary>
        /// Use CloseSession() to end session. If possible, verify that the following service calls to the session are rejected.
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group001_Session_Test06_CloseSession()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient {Isolated = true})
            {
                var hasDisconnected = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
                var hasFailed = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
                client.ServerConditionChanged += (sender, args) =>
                {
                    Debug.Assert(args != null);
                    if ((args.Connected == false) && (args.Exception == null))
                        hasDisconnected.Set();
                    if ((args.Connected == false) && (args.Exception != null))
                        hasFailed.Set();
                };

                var nodeDescriptor = new UANodeDescriptor(UAVariableIds.Server_ServerCapabilities_MaxArrayLength);
                try
                {
                    var value = client.ReadValue(endpointDescriptor, nodeDescriptor);
                    AssertUtilities.WriteLine($"Value: {value}");
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine($"UA Exception: {uaException}");
                }
                var handleIndex = WaitHandle.WaitAny(new WaitHandle[] { hasDisconnected, hasFailed }, 10 * 1000);
                Assert.AreEqual(0, handleIndex, "Has not successfully disconnected");
            }
        }

        /// <summary>
        /// Create and use multiple sessions.
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test07_CreateAndUseMultipleSessions()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client1 = new EasyUAClient { Isolated = true })
            using (var client2 = new EasyUAClient { Isolated = true })
            {
                bool isSuccessfullyConnected1 = false;
                bool isSuccessfullyConnected2 = false;
                var bothConnected = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
                var hasFailed = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
                var synchronize = new object();

                void Handler(object sender, EasyUAServerConditionChangedEventArgs args)
                {
                    Debug.Assert(args != null);
                    lock (synchronize)
                    {
                        if (args.Exception == null)
                        {
                            // ReSharper disable AccessToDisposedClosure
                            if (ReferenceEquals(sender, client1))
                                isSuccessfullyConnected1 = args.Connected;
                            else if (ReferenceEquals(sender, client2)) isSuccessfullyConnected2 = args.Connected;
                            // ReSharper restore AccessToDisposedClosure
                        }

                        if (isSuccessfullyConnected1 && isSuccessfullyConnected2) bothConnected.Set();
                        if ((args.Connected == false) && (args.Exception != null)) hasFailed.Set();
                    }
                }

                client1.ServerConditionChanged += Handler;
                client2.ServerConditionChanged += Handler;

                var nodeDescriptor = new UANodeDescriptor(UAVariableIds.Server_ServerCapabilities_MaxArrayLength);
                try
                {
                    UAAttributeData attributeData1 = client1.Read(endpointDescriptor, nodeDescriptor);
                    AssertUtilities.WriteLine($"Attribute Data 1: {attributeData1}");

                    AssertUtilities.WaitingPhase(InterSessionDelay);

                    UAAttributeData attributeData2 = client2.Read(endpointDescriptor, nodeDescriptor);
                    AssertUtilities.WriteLine($"Attribute Data 2: {attributeData2}");
                }
                catch (UAException uaException)
                {
                    AssertUtilities.WriteLine($"UA Exception: {uaException}");
                }

                TestSection.Assert();

                int handleIndex = WaitHandle.WaitAny(new WaitHandle[] { bothConnected, hasFailed }, 10 * 1000);
                Assert.AreNotEqual(WaitHandle.WaitTimeout, handleIndex, "None failed but both have not connected");
                Assert.AreEqual(0, handleIndex, "At least one has not successfully connected");
            }
        }

        /// <summary>
        /// CreateSession() with an application instance certificate
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test08_CreateSession_InstanceCertificate()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
                { EndpointSelectionPolicy = new UAEndpointSelectionPolicy(UAMessageSecurityModes.Secure) };

            TestSection.Act();
            using (var client = new EasyUAClient {Isolated = true})
            {
                var nodeDescriptor = new UANodeDescriptor(UAVariableIds.Server_ServerCapabilities_MaxArrayLength);
                UAAttributeData attributeData = client.Read(endpointDescriptor, nodeDescriptor);
                AssertUtilities.WriteLine($"Attribute data: {attributeData}");
            }
        }

        /// <summary>
        /// CreateSession() over a secure channel with HTTPS and SecurityPolicy =None.
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test09_CreateSession_Https_None()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorHttps = EndpointDescriptorHttps;
            Assert.IsNotNull(endpointDescriptorHttps, "No HTTPS endpoint descriptor given");
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorHttps))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorHttps)
                { EndpointSelectionPolicy = new UAEndpointSelectionPolicy(UAMessageSecurityModes.SecurityNone) };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// ActivateSession() with user identity = X509 certificate
        /// </summary>
        [Ignore]    // we no longer "advertise" the support for this during IOP WS
        [TestMethod]
        public void Group001_Session_Test10_ActivateSession_X509Certificate()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            string x509CertificateFileName = X509CertificateFileName;
            Debug.Assert(x509CertificateFileName != null, "Missing X509 certificate file name");
            var endpointDescriptor = new UAEndpointDescriptor(EndpointDescriptorCommon)
                { UserIdentity = UserIdentity.CreateX509CertificateIdentity(x509CertificateFileName, X509CertificatePassword) };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// CreateSession() over a secure channel with UA TCP and SecurityPolicy=Basic256Sha256
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test11_CreateSession_UaTcp_Basic256Sha256()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
            {
                EndpointSelectionPolicy = new UAEndpointSelectionPolicy { SecurityPolicyUriString = UASecurityPolicyUriStrings.Basic256Sha256 }
            };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        // ReSharper disable once CommentTypo
        /// <summary>
        /// CreateSession() over a secure channel with UA TCP and SecurityPolicy=Aes128-Sha256-RsaOaep
        /// </summary>
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void Group001_Session_Test12_CreateSession_UaTcp_Aes128_Sha256_RsaOaep()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
            {
                EndpointSelectionPolicy = new UAEndpointSelectionPolicy
                {
                    SecurityPolicyUriString = UASecurityPolicyUriStrings.Aes128_Sha256_RsaOaep
                }
            };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        /// <summary>
        /// CreateSession() over a secure channel with UA TCP and SecurityPolicy=Aes256-Sha256-RsaPss
        /// </summary>
        [TestMethod]
        public void Group001_Session_Test13_CreateSession_UaTcp_Aes256_Sha256_RsaPss()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
            Assert.IsNotNull(endpointDescriptorUATcp);
            if (!UALiveTestEnvironment.IsSupportedOptimistic(endpointDescriptorUATcp))
                ClientNotSupported();
            var endpointDescriptor = new UAEndpointDescriptor(endpointDescriptorUATcp)
            {
                EndpointSelectionPolicy = new UAEndpointSelectionPolicy
                {
                    SecurityPolicyUriString = UASecurityPolicyUriStrings.Aes256_Sha256_RsaPss
                }
            };

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, endpointDescriptor);
        }

        #endregion



        #region Private Methods

        private void CreateSession([NotNull] EasyUAClient client, [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var hasConnected = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
            var hasFailed = new EventWaitHandle(initialState: false, mode: EventResetMode.ManualReset);
            client.ServerConditionChanged += (sender, args) =>
            {
                Debug.Assert(args != null);
                AssertUtilities.WriteLine($"Server condition changed: {args}");
                if ((args.Connected) && (args.Exception == null))
                    hasConnected.Set();
                if ((!args.Connected) && (args.Exception != null))
                    hasFailed.Set();
            };

            var nodeDescriptor = new UANodeDescriptor(UAVariableIds.Server_ServerCapabilities_MaxArrayLength);
            try
            {
                AssertUtilities.WriteLine("Reading...");
                UAAttributeData attributeData = client.Read(endpointDescriptor, nodeDescriptor);
                AssertUtilities.WriteLine($"Attribute data: {attributeData}");
            }
            catch (UAException uaException)
            {
                AssertUtilities.WriteLine($"UA Exception: {uaException}");
                // do not rethrow
            }

            int handleIndex = WaitHandle.WaitAny(new WaitHandle[] { hasConnected, hasFailed }, 10 * 1000);
            Assert.AreEqual(0, handleIndex, "Has not successfully connected");
        }

        #endregion
    }
}
